==> mars-0.9.2.tar.gz: mars-0.9.2/src/host/lib/mars.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <malloc.h>
#include <string.h>
#include <pthread.h>
#include <libspe2.h>

#include "mars/mars.h"
#include "mars/mars_workload_queue.h"
#include "mars/mars_kernel_types.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

extern struct spe_program_handle mars_kernel_entry;

static unsigned int num_mpus_max(void)
{
	return spe_cpu_info_get(SPE_COUNT_PHYSICAL_SPES, -1);
}

static void *mpu_context_thread(void *arg)
{
	MARS_ASSERT(arg);

	int ret;
	unsigned int entry = SPE_DEFAULT_ENTRY;
	struct mars_kernel_params *params = (struct mars_kernel_params *)arg;
	struct spe_context *spe;

	spe = spe_context_create(0, NULL);
	MARS_CHECK_RET(spe, (void *)MARS_ERROR_INTERNAL);

	ret = spe_program_load(spe, &mars_kernel_entry);
	MARS_CHECK_RET(!ret, (void *)MARS_ERROR_INTERNAL);

	ret = spe_context_run(spe, &entry, 0, params, NULL, NULL);
	MARS_CHECK_RET(!ret, (void *)MARS_ERROR_INTERNAL);

	ret = spe_context_destroy(spe);
	MARS_CHECK_RET(!ret, (void *)MARS_ERROR_INTERNAL);

	return (void *)MARS_SUCCESS;
}

static int create_mpu_contexts(struct mars_context *mars,
				unsigned int num_mpu_contexts)
{
	MARS_ASSERT(mars);

	int ret;
	unsigned int i;

	/* create threads for each mpu context */
	for (i = 0; i < num_mpu_contexts; i++) {
		struct mars_kernel_params *params = &mars->kernel_params[i];
		params->id = i;
		params->mars_context_ea =
				(uint64_t)(uintptr_t)mars;
		params->workload_queue_ea =
				(uint64_t)(uintptr_t)mars->workload_queue;

		ret = pthread_create(&mars->mpu_context_threads[i], NULL,
				mpu_context_thread, params);
		MARS_CHECK_RET(!ret, MARS_ERROR_INTERNAL);

		mars->mpu_context_count++;
	}

	return MARS_SUCCESS;
}

static int destroy_mpu_contexts(struct mars_context *mars)
{
	MARS_ASSERT(mars);

	int ret;
	void *p_ret;
	unsigned int i;

	/* join all mpu context threads */
	for (i = 0; i < mars->mpu_context_count; i++) {
		ret = pthread_join(mars->mpu_context_threads[i], &p_ret);
		MARS_CHECK_RET(!ret && !p_ret, MARS_ERROR_INTERNAL);
	}

	return MARS_SUCCESS;
}

int mars_initialize(struct mars_context *mars, struct mars_params *params)
{
	MARS_CHECK_RET(mars, MARS_ERROR_NULL);

	MARS_PRINT("Initialize MARS Context (%p)\n", mars);

	int ret;
	unsigned int num_mpu_contexts;

	/* zero context */
	memset(mars, 0, sizeof(struct mars_context));

	/* allocate and initialize params */
	if (params) {
		MARS_CHECK_RET(params->num_mpus <= num_mpus_max(),
				MARS_ERROR_PARAMS);

		mars->params = (struct mars_params *)
			malloc(sizeof(struct mars_params));
		MARS_CHECK_RET(mars->params, MARS_ERROR_MEMORY);
		memcpy(mars->params, params, sizeof(struct mars_params));

		num_mpu_contexts = mars->params->num_mpus;
	} else {
		mars->params = NULL;

		num_mpu_contexts = num_mpus_max();
	}

	/* allocate kernel params */
	mars->kernel_params = (struct mars_kernel_params *)
		memalign(MARS_KERNEL_PARAMS_ALIGN,
			sizeof(struct mars_kernel_params) * num_mpu_contexts);
	MARS_CHECK_CLEANUP_RET(mars->kernel_params, mars_finalize(mars),
				MARS_ERROR_MEMORY);

	/* allocate workload queue */
	mars->workload_queue = (struct mars_workload_queue *)
		memalign(MARS_WORKLOAD_QUEUE_ALIGN,
			sizeof(struct mars_workload_queue));
	MARS_CHECK_CLEANUP_RET(mars->workload_queue, mars_finalize(mars),
				MARS_ERROR_MEMORY);

	/* initialize workload queue */
	ret = workload_queue_initialize(mars->workload_queue);
	MARS_CHECK_CLEANUP_RET(ret == MARS_SUCCESS, mars_finalize(mars), ret);

	/* allocate mpu context thread array */
	mars->mpu_context_threads = (pthread_t *)
		malloc(sizeof(pthread_t) * num_mpu_contexts);
	MARS_CHECK_CLEANUP_RET(mars->mpu_context_threads, mars_finalize(mars),
				MARS_ERROR_MEMORY);

	/* create contexts */
	ret = create_mpu_contexts(mars, num_mpu_contexts);
	MARS_CHECK_CLEANUP_RET(ret == MARS_SUCCESS, mars_finalize(mars), ret);

	return MARS_SUCCESS;
}

int mars_finalize(struct mars_context *mars)
{
	MARS_CHECK_RET(mars, MARS_ERROR_NULL);

	int ret;

	/* finalize workload queue */
	if (mars->workload_queue) {
		ret = workload_queue_finalize(mars->workload_queue);
		MARS_CHECK_RET(ret == MARS_SUCCESS, ret);
	}

	/* destroy contexts */
	if (mars->mpu_context_count) {
		ret = destroy_mpu_contexts(mars);
		MARS_CHECK_RET(ret == MARS_SUCCESS, ret);
	}

	/* free allocated memory */
	free(mars->params);
	free(mars->kernel_params);
	free(mars->workload_queue);
	free(mars->mpu_context_threads);

	/* zero context */
	memset(mars, 0, sizeof(struct mars_context));

	MARS_PRINT("Finalize MARS Context (%p)\n", mars);

	return MARS_SUCCESS;
}

==> mars-08.06.26.tar.gz: mars-08.06.26/src/host/lib/mars.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <malloc.h>
#include <string.h>
#include <pthread.h>
#include <libspe2.h>

#include "mars/mars.h"
#include "mars/mars_workload_queue.h"
#include "mars/mars_kernel_types.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

extern struct spe_program_handle mars_kernel_entry;

static int num_mpus_max(void)
{
	return spe_cpu_info_get(SPE_COUNT_PHYSICAL_SPES, 0);
}

static void *mpu_context_thread(void *arg)
{
	MARS_ASSERT(arg);

	int ret;
	unsigned int entry = SPE_DEFAULT_ENTRY;
	struct mars_kernel_params *params = (struct mars_kernel_params *)arg;
	struct spe_context *spe;

	spe = spe_context_create(0, NULL);
	MARS_CHECK_CLEANUP_RET(spe, ret = MARS_ERROR_INTERNAL;
				pthread_exit(&ret), NULL);

	ret = spe_program_load(spe, &mars_kernel_entry);
	MARS_CHECK_CLEANUP_RET(!ret, ret = MARS_ERROR_INTERNAL;
				pthread_exit(&ret), NULL);

	ret = spe_context_run(spe, &entry, 0,
				params, NULL, NULL);
	MARS_CHECK_CLEANUP_RET(!ret, ret = MARS_ERROR_INTERNAL;
				pthread_exit(&ret), NULL);

	ret = spe_context_destroy(spe);
	MARS_CHECK_CLEANUP_RET(!ret, ret = MARS_ERROR_INTERNAL;
				pthread_exit(&ret), NULL);

	pthread_exit(MARS_SUCCESS);

	return NULL;
}

static int create_mpu_contexts(struct mars_context *mars, int num_mpu_contexts)
{
	MARS_ASSERT(mars);

	int ret;
	unsigned int i;

	/* create threads for each mpu context */
	for (i = 0; i < num_mpu_contexts; i++) {
		struct mars_kernel_params *params = &mars->kernel_params[i];
		params->id = i;
		params->mars_context_ea =
				(uint64_t)(uintptr_t)mars;
		params->workload_queue_ea =
				(uint64_t)(uintptr_t)mars->workload_queue;

		ret = pthread_create(&mars->mpu_context_threads[i], NULL,
				mpu_context_thread, params);
		MARS_CHECK_RET(!ret, MARS_ERROR_INTERNAL);

		mars->mpu_context_count++;
	}

	return MARS_SUCCESS;
}

static int destroy_mpu_contexts(struct mars_context *mars)
{
	MARS_ASSERT(mars);

	int ret;
	int thread_ret;
	void *ptr = &thread_ret;
	unsigned int i;

	/* join all mpu context threads */
	for (i = 0; i < mars->mpu_context_count; i++) {
		ret = pthread_join(mars->mpu_context_threads[i], &ptr);
		MARS_CHECK_RET(!ret, thread_ret);
	}

	return MARS_SUCCESS;
}

int mars_initialize(struct mars_context *mars, struct mars_params *params)
{
	MARS_CHECK_RET(mars, MARS_ERROR_NULL);

	MARS_PRINT("Initialize MARS Context (%p)\n", mars);

	int ret;
	unsigned int num_mpu_contexts;

	/* zero context */
	memset(mars, 0, sizeof(struct mars_context));

	/* allocate and initialize params */
	if (params) {
		MARS_CHECK_RET(params->num_mpus <= num_mpus_max(),
				MARS_ERROR_PARAMS);

		mars->params = (struct mars_params *)
			malloc(sizeof(struct mars_params));
		MARS_CHECK_RET(mars->params, MARS_ERROR_MEMORY);
		memcpy(mars->params, params, sizeof(struct mars_params));

		num_mpu_contexts = mars->params->num_mpus;
	} else {
		mars->params = NULL;

		num_mpu_contexts = num_mpus_max();
	}

	/* allocate kernel params */
	mars->kernel_params = (struct mars_kernel_params *)
		memalign(MARS_KERNEL_PARAMS_ALIGN,
			sizeof(struct mars_kernel_params) * num_mpu_contexts);
	MARS_CHECK_CLEANUP_RET(mars->kernel_params, mars_finalize(mars),
				MARS_ERROR_MEMORY);

	/* allocate mutex */
	mars->mutex = (struct mars_mutex *)
		memalign(MARS_MUTEX_ALIGN, sizeof(struct mars_mutex));
	MARS_CHECK_CLEANUP_RET(mars->mutex, mars_finalize(mars),
				MARS_ERROR_MEMORY);

	/* initialize mutex */
	ret = mars_mutex_initialize(mars->mutex);
	MARS_CHECK_CLEANUP_RET(ret == MARS_SUCCESS, mars_finalize(mars), ret);

	/* allocate workload queue */
	mars->workload_queue = (struct mars_workload_queue *)
		memalign(MARS_WORKLOAD_QUEUE_ALIGN,
			sizeof(struct mars_workload_queue));
	MARS_CHECK_CLEANUP_RET(mars->workload_queue, mars_finalize(mars),
				MARS_ERROR_MEMORY);

	/* initialize workload queue */
	ret = workload_queue_initialize(mars->workload_queue);
	MARS_CHECK_CLEANUP_RET(ret == MARS_SUCCESS, mars_finalize(mars), ret);

	/* allocate mpu context thread array */
	mars->mpu_context_threads = (pthread_t *)
		malloc(sizeof(pthread_t) * num_mpu_contexts);
	MARS_CHECK_CLEANUP_RET(mars->mpu_context_threads, mars_finalize(mars),
				MARS_ERROR_MEMORY);

	/* create contexts */
	ret = create_mpu_contexts(mars, num_mpu_contexts);
	MARS_CHECK_CLEANUP_RET(ret == MARS_SUCCESS, mars_finalize(mars), ret);

	return MARS_SUCCESS;
}

int mars_finalize(struct mars_context *mars)
{
	MARS_CHECK_RET(mars, MARS_ERROR_NULL);

	int ret;

	/* finalize workload queue */
	if (mars->workload_queue) {
		ret = workload_queue_finalize(mars->workload_queue);
		MARS_CHECK_RET(ret == MARS_SUCCESS, ret);
	}

	/* destroy contexts */
	if (mars->mpu_context_count) {
		ret = destroy_mpu_contexts(mars);
		MARS_CHECK_RET(ret == MARS_SUCCESS, ret);
	}

	/* free allocated memory */
	free(mars->params);
	free(mars->kernel_params);
	free(mars->mutex);
	free(mars->workload_queue);
	free(mars->mpu_context_threads);

	/* zero context */
	memset(mars, 0, sizeof(struct mars_context));

	MARS_PRINT("Finalize MARS Context (%p)\n", mars);

	return MARS_SUCCESS;
}
